#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <time.h>

// Remember, operator mutation is half of this. 
#define MUTATE_PROB 0.05
#define NUM_LOGIC_GATES 16

typedef uint32_t state_t;

typedef enum {		// 00 01 10 11
	FALSE, 			// 0  0  0  0
	AND,			// 0  0  0  1
	MAT_NONIMPL,	// 0  0  1  0
	A_PROJ,			// 0  0  1  1
	CONV_NONIMPL,	// 0  1  0  0
	B_PROJ,			// 0  1  0  1
	XOR,			// 0  1  1  0	
	OR, 			// 0  1  1  1	 
	NOR, 			// 1  0  0  0
	XNOR,			// 1  0  0  1
	NOT_B,			// 1  0  1  0
	CONV_IMPL,		// 1  0  1  1
	NOT_A,			// 1  1  0  0
	MAT_IMPL,		// 1  1  0  1
	NAND,			// 1  1  1  0
	TRUE			// 1  1  1  1
} gate_t;

typedef struct {
	gate_t op1;
	gate_t op2;
	uint32_t in_A;
	uint32_t in_B;
	uint32_t in_C;
} node_t;

typedef enum {VISITED, NOT_VISITED} Visited;

node_t *network_create (int num_nodes) {
	node_t *network = malloc(num_nodes*sizeof(node_t));
	for (int i = 0; i < num_nodes; i++) {
		network[i].op1 = (gate_t)(rand() % NUM_LOGIC_GATES);
		network[i].op2 = (gate_t)(rand() % NUM_LOGIC_GATES);
		network[i].in_A = rand() % num_nodes;
		network[i].in_B = rand() % num_nodes;
		network[i].in_C = rand() % num_nodes;
	}
	return network;
}

void network_destroy (node_t *network) {
	free(network);
}

//New state calculation + pointer conversion.
state_t ptr_to_state (state_t *state, int num_nodes) {
	int ret = 0;
	for (int i = 0; i < num_nodes; i++) {
		ret |= (state[i])<<i;
	}
	return ret;
}

state_t *state_to_ptr (state_t state, int num_nodes) {
	state_t *ret_ptr = malloc(sizeof(state_t) * num_nodes);
	for (int i = 0; i < num_nodes; i++) {
		ret_ptr[i] = (state>>i)&1;
	}
	return ret_ptr;
}

state_t eval(gate_t operator, state_t a, state_t b) {
	switch (operator) {
		case FALSE:
			return 0;
		case AND:
			return a & b;
		case MAT_NONIMPL:
			if (a == 1 && b == 0) return 1;
			return 0;
		case A_PROJ:
			return a;
		case CONV_NONIMPL:
			if (a == 0 && b == 1) return 1;
			return 0;
		case B_PROJ:
			return b;
		case XOR:
			return a ^ b;
		case OR:
			return a | b;
		case NOR:
			return !(a | b);
		case XNOR:
			return !(a ^ b);
		case NOT_B:
			return !b;
		case CONV_IMPL:
			if (a == 0 && b == 1) return 0;
			return 1;
		case NOT_A:
			return !a;
		case MAT_IMPL:
			if (a == 1 && b == 0) return 0;
			return 1;
		case NAND:
			return !(a & b);
		case TRUE:
			return 1;
	}
	return -1; // Just in case.
}

state_t calc_new_state(node_t *network, state_t *prev_state, int node_index) {
	gate_t operator1 = network[node_index].op1;
	gate_t operator2 = network[node_index].op2;
	state_t a = prev_state[network[node_index].in_A];
	state_t b = prev_state[network[node_index].in_B];
	state_t c = prev_state[network[node_index].in_C];
	return eval(operator2, eval(operator1, a, b), c);
}

state_t next_state (node_t *network, state_t previous_state, int num_nodes) {
	state_t *next_state = malloc(sizeof(state_t) * num_nodes);
	state_t *prev_state = state_to_ptr(previous_state, num_nodes);
	for (int i = 0; i < num_nodes; i++) {
		next_state[i] = calc_new_state(network, prev_state, i);
		if (next_state[i] == -1) {
			printf("Illegal node operator. Terminating.\n");
			exit(1);
		}
	}
	state_t result = ptr_to_state(next_state, num_nodes);
	free(prev_state);
	free(next_state);
	return result;
}

// Print functions for feedback.
void operator_print(gate_t op, FILE * file) {
	switch (op) {
		case FALSE:
			fprintf(file,"FALSE");
			break;
		case AND:
			fprintf(file,"AND");
			break;
		case MAT_NONIMPL:
			fprintf(file,"MAT_NONIMPL");
			break;
		case A_PROJ:
			fprintf(file,"A_PROJ");
			break;
		case CONV_NONIMPL:
			fprintf(file,"CONV_NONIMPL");
			break;
		case B_PROJ:
			fprintf(file,"B_PROJ");
			break;
		case XOR:
			fprintf(file,"XOR");
			break;
		case OR:
			fprintf(file,"OR");
			break;
		case NOR:
			fprintf(file,"NOR");
			break;
		case XNOR:
			fprintf(file,"XNOR");
			break;
		case NOT_B:
			fprintf(file,"NOT_B");
			break;
		case CONV_IMPL:
			fprintf(file,"CONV_IMPL");
			break;
		case NOT_A:
			fprintf(file,"NOT_A");
			break;
		case MAT_IMPL:
			fprintf(file,"MAT_IMPL");
			break;;
		case NAND:
			fprintf(file,"NAND");
			break;
		case TRUE:
			fprintf(file,"TRUE");
			break;
	}
}

void array_print(node_t *network, int num_nodes, FILE * file) {
	fprintf(file, "{");
	for (int i = 0; i < num_nodes-1; i++) {
		operator_print(network[i].op1, file);
		fprintf(file, "-");
		operator_print(network[i].op2, file);
		fprintf(file,", ");
	}
	operator_print(network[num_nodes-1].op1, file);
	fprintf(file, "-");
	operator_print(network[num_nodes-1].op2, file);
	fprintf(file,"} \n");
	
	fprintf(file, "{");
	for (int i = 0; i < num_nodes-1; i++) {
		fprintf(file,"%d, ", network[i].in_A);
	}
	fprintf(file,"%d} \n", network[num_nodes-1].in_A);
	
	fprintf(file,"{");
	for (int i = 0; i < num_nodes-1; i++) {
		fprintf(file, "%d, ", network[i].in_B);
	}
	fprintf(file,"%d} \n", network[num_nodes-1].in_B);
	fprintf(file,"{");
	for (int i = 0; i < num_nodes-1; i++) {
		fprintf(file, "%d, ", network[i].in_C);
	}
	fprintf(file,"%d} \n", network[num_nodes-1].in_C);
}

// Generating random states.
float randomFloat() {
      float r = (float)rand()/(float)RAND_MAX;
      return r;
}

void shuffle(state_t *array, size_t n) {
    if (n > 1) {
        size_t i;
        for (i = 0; i < n - 1; i++) {
          size_t j = i + rand() / (RAND_MAX / (n - i) + 1);
          state_t t = array[j];
          array[j] = array[i];
          array[i] = t;
        }
    }
}

state_t *generate_random_states(int num_nodes) {
	int num_states = 1<<num_nodes;
	state_t * result = malloc(sizeof(state_t)*num_states);	
	for (int i = 0; i < num_states; i++) {
		result[i] = i;
	}
	shuffle(result, num_states);
	return result;
}

node_t *network_mutate(node_t *network, int num_nodes){
	node_t *result = malloc(num_nodes*sizeof(node_t));
	for (int i = 0; i < num_nodes; i++) {
		result[i].op1 = network[i].op1;
		result[i].op2 = network[i].op2;
		result[i].in_A = network[i].in_A;
		result[i].in_B = network[i].in_B;
		result[i].in_C = network[i].in_C;
	}
	for (int i = 0; i < num_nodes; i++) {
		if (MUTATE_PROB/2 > randomFloat()) {
			result[i].op1 = (gate_t)(rand() % NUM_LOGIC_GATES);
		}
		if (MUTATE_PROB/2 > randomFloat()) {
			result[i].op2 = (gate_t)(rand() % NUM_LOGIC_GATES);
		}
		if (MUTATE_PROB > randomFloat()) {
			result[i].in_A = rand() % num_nodes;
		}
		if (MUTATE_PROB > randomFloat()) {
			result[i].in_B = rand() % num_nodes;
		}
		if (MUTATE_PROB > randomFloat()) {
			result[i].in_C = rand() % num_nodes;
		}
	}
	return result;
}

void state_history_save(state_t *state_history, Visited *visited, int iteration) {
	for (int i = 0; i < iteration; i++) {
		visited[state_history[i]] = VISITED;
	}
	free(state_history);
}

int network_fitness(node_t *network, int num_nodes, int sample_divisor) {
	int num_states = 1<<num_nodes;
	int max_ticks = num_states+1;
	int max_attractor_len = 0;
	state_t * init_states = generate_random_states(num_nodes);
	Visited *visited = malloc(sizeof(Visited)*num_states);
	for (int i = 0; i < num_states; i++) {
		visited[i] = NOT_VISITED;
	}

	for (int i = 0; i < num_states/sample_divisor; i++) {
		state_t start_state = init_states[i];
		
		if (visited[start_state] == VISITED) {
			goto endcycle;	
		}
		state_t *state_history = malloc(sizeof(state_t)*num_states);
		state_history[0] = start_state;
		state_t new_state;
		//Check whether start_state is visited.
		int j = 0;
		while (j < max_ticks) {
			j++;
			new_state = next_state(network, state_history[j-1], num_nodes);
			if (visited[new_state] == VISITED) {
				state_history_save(state_history, visited, j);
				break;
			}
			if (memcmp(&state_history[j-1], &new_state, sizeof(state_t)) == 0) {
				if (max_attractor_len < 1) {
					max_attractor_len = 1;
				}
				state_history_save(state_history, visited, j);
				break;
			}
			else {
				for (int k = 0; k < j; k++) { // Cycle through history, starting at 0 because it breaks the loop on first detected cycle anyway.
					if (memcmp(&state_history[k], &new_state, sizeof(state_t)) == 0) {
						if (max_attractor_len < (j-k)) {
						max_attractor_len = (j-k);
						}
						state_history_save(state_history, visited, j);
						goto endcycle; // Because break doesn't break out of the while loop.
					}
				}
			}
			state_history[j] = new_state;
		}
		endcycle: ; //Labels are associated with statements and all that.
	}
	free(init_states);
	free(visited);
	return max_attractor_len;
}

node_t * network_reproduce(node_t *parent_one, node_t *parent_two, int num_nodes) {
	node_t * child = malloc(num_nodes*sizeof(node_t));
	int crossover_point = rand() % num_nodes;
	for (int i = 0; i < crossover_point; i++) {
		child[i] = parent_one[i];
	}
	for (int j = crossover_point; j < num_nodes; j++) {
		child[j] = parent_two[j];
	}
	node_t * temp = network_mutate(child, num_nodes);
	free(child);
	child = temp;
	return child;
}

void nextBest(int *arr, int n, int *max, int *maxIndex) {
  int cand = -1;
  int candIndex = -1;
  for (int i = 0; i < n; i++) {
    // max cand < arr[i] < cur max
    if ((cand < arr[i]) && (arr[i] < *max)) {
      cand = arr[i];
      candIndex = i;
      continue;
    }
    // arr[i] == max, and maxIndex < i
    if ((cand < arr[i]) && (arr[i] == *max) && (*maxIndex < i)) {
      cand = arr[i];
      candIndex = i;
      continue;
    }
  }
  *max = cand;
  *maxIndex = candIndex;
}

int main(int argc, char *argv[]) {
	if (argc < 7) {
		printf("Too few arguments. Arguments needed: \n Stat file name, graph file name, # nodes, # generations, # population, # elites\n");
		exit(1);
	}
	unsigned int seed; 
	FILE* urandom = fopen("/dev/urandom", "r");
	fread(&seed, sizeof(int), 1, urandom);
	fclose(urandom); 
	srand(seed);
	FILE *stats;
	FILE *graph;
	char *statsfile = argv[1];
	char *graphfile = argv[2];
	int num_nodes = atoi(argv[3]);
	int total_gens = atoi(argv[4]);
	int pop_count = atoi(argv[5]);
	int elite_num = atoi(argv[6]);
	
	stats = fopen (statsfile, "w");
	graph = fopen (graphfile, "w");
	
	int total_fitness = 0;
	int sample_divisor = 1; // For reducing the number of states tested.
	int fitness_val = 0;
	int pi1 = 0; // Parent indexes.
	int pi2 = 0;
	int rand1 = 0; // Roulette variables.
	int rand2 = 0;
	int max_fitness;
	int min_fitness;
	double mean_fitness;
	double sd;
	double temp;
	int max; // For elitism.
	int maxIndex;
	
	struct timespec time_start, time_stop;
	long long nanoseconds;

	clock_gettime(CLOCK_MONOTONIC, &time_start);
	
	node_t ** population_array = malloc(sizeof(node_t*)*pop_count);
	int * fitness_array = malloc(sizeof(int)*pop_count);	
	for (int i = 0; i < pop_count; i++) {
			population_array[i] = network_create(num_nodes);
		}
		
	//Actual evolution comes here:
	for (int i = 0; i < total_gens; i++) {
		node_t ** child_array = malloc(sizeof(node_t*)*pop_count);
		
		// Calculating MAX, MIN, MEAN, SD.
		max_fitness = 0;
		min_fitness = 1<<num_nodes;
		mean_fitness = 0.0;
		sd = 0.0;
		temp = 0.0;
		total_fitness = 0;
		
		for (int j = 0; j < pop_count; j++) {
			fitness_val = network_fitness(population_array[j], num_nodes, sample_divisor);
			total_fitness += fitness_val;
			if (fitness_val < min_fitness) min_fitness = fitness_val;
			if (fitness_val > max_fitness) max_fitness = fitness_val;
			fitness_array[j] = fitness_val;
		}
		
		mean_fitness = (double)total_fitness / (double)pop_count;
		for (int o = 0; o < pop_count; o++) {
			temp += pow((fitness_array[o]-mean_fitness), 2.0);
		}
		sd = sqrt(temp/((double)pop_count));
		//Format: GEN MAX MIN MEAN SD
		fprintf(stats, "%d %d %d %.5f %.5f \n", i, max_fitness, min_fitness, mean_fitness, sd);
		
		max = total_fitness;
		maxIndex = -1;
		// Network reproduction happens here. Roulette wheel strategy + elitism
		for (int k = 0; k < elite_num; k++) {
			nextBest(fitness_array, pop_count, &max, &maxIndex);
			node_t * elite = malloc(num_nodes*sizeof(node_t));
			memcpy(elite, population_array[maxIndex], (num_nodes*sizeof(node_t)));
			child_array[k] = elite;
		}
		
		for (int k = elite_num; k < pop_count; k++) {
			rand1 = rand() % total_fitness;
			rand2 = rand() % total_fitness;
			for (int m = 0; m < pop_count; m++) {
				if (fitness_array[m] < rand1) {
					pi1 = m;
					break;
				}
				else {
					rand1 -= fitness_array[m];
				}
			}
			for (int n = 0; n < pop_count; n++) {
				if (fitness_array[n] < rand2) {
					pi2 = n;
					break;
				}
				else {
					rand2 -= fitness_array[n];
				}
			}
			child_array[k] = network_reproduce(population_array[pi1], population_array[pi2], num_nodes);	
		}
		
		for(int i = 0; i < pop_count; i++) {
			free(population_array[i]);
		}
		free(population_array);
		population_array = child_array;
	}
	
	// LAST GENERATION
	max_fitness = 0;
	min_fitness = 1<<num_nodes;
	mean_fitness = 0.0;
	sd = 0.0;
	temp = 0.0;
	total_fitness = 0;
	
	for (int j = 0; j < pop_count; j++) {
		fitness_val = network_fitness(population_array[j], num_nodes, sample_divisor);
		total_fitness += fitness_val;
		if (fitness_val < min_fitness) min_fitness = fitness_val;
		if (fitness_val > max_fitness) max_fitness = fitness_val;
		fitness_array[j] = fitness_val;
	}
	
	mean_fitness = (double)total_fitness / (double)pop_count;
	for (int o = 0; o < pop_count; o++) {
		temp += pow((fitness_array[o]-mean_fitness), 2.0);
	}
	sd = sqrt(temp/((double)pop_count));
	//Format: GEN MAX MIN MEAN SD
	fprintf(stats, "%d %d %d %.5f %.5f \n", total_gens, max_fitness, min_fitness, mean_fitness, sd);
	
	// Printing the best network(s) after M generations!
	for (int i = 0; i < pop_count; i++) {
		if (fitness_array[i] == max_fitness) {
			//network_print(population_array[i], num_nodes);
			array_print(population_array[i], num_nodes, graph);
			break;
		}
	}
	fclose(stats);
	fclose(graph);
	for(int i = 0; i < pop_count; i++) {
		free(population_array[i]);
	}
	free(population_array);
	free(fitness_array);
	clock_gettime(CLOCK_MONOTONIC, &time_stop);
	nanoseconds = (time_stop.tv_sec - time_start.tv_sec) * 1000000000LL;
	nanoseconds += (time_stop.tv_nsec - time_start.tv_nsec);
	printf("Time: %lld \n", nanoseconds);
	return 0;
}
